# Causal Consistency

## Causal Basic

### Causal Basic 1: Bob Smells

#### Test Description
The classic "bob smells" example. This test involves three clients and two server nodes. The first client writes x to a server node. The second client reads x from the same server node and writes y to the other server node. The last client performs two reads of y and x, in that order, to the latter server node. This server should hang.

#### Explicit Violation
A violation would be if the last client reads anything from the server and terminates.

#### Lamport Diagram
![Lamport Diagram - Bob Smells Example](./lamport_bob_smells.jpg)

### Causal Basic 1: Bob Smells

#### Test Description
The classic "bob smells" example. This test involves three clients and two server nodes. The first client writes x to a server node. The second client reads x from the same server node and writes y to the other server node. The last client performs two reads of y and x, in that order, to the latter server node. This server should hang.

#### Explicit Violation
A violation would be if the last client reads anything from the server and terminates.

#### Lamport Diagram
![Lamport Diagram - Bob Smells Example](./lamport_causal_bob_smells.jpg)

### Causal Basic 2: Healing Bob Smells

#### Test Description
The classic "bob smells" example. This test involves three clients and two server nodes. The first client writes x to a server node. The second client reads x from the same server node and writes y to the other server node. The last client performs two reads of y and x, in that order, to the latter server node. This server should first hang. Then after the partition has healed, it should be able to read the correct value.

#### Explicit Violation
A violation would be if the server hangs forever of if the last read returns something other than "bob smells".

#### Lamport Diagram
![Lamport Diagram - Healing Bob Smells Example](./lamport_causal_heal_bob_smells.jpg)

### Causal advanced: Bob doesn't smell(old)
This test involves 3 clients and 2 server nodes, and is a play on the traditional bob smells test. The test involves two stages, the first, where the two servers are partitioned, and the second, where the two servers can talk to each other. From now on, we will refer to the clients as ca, cb, cc, and the servers as s0, s1. The only possible keys are x and y. Note: the passing of this test *implicitly* expects the server to tiebreak using timestamps after the primary tiebreaking method, and this may fail if a user has implemented a different tiebreaking method.

The procedure is as follows:

- ca put x in s0 (should be ok)
- ca read x from s1 (should hang)
- cb read x from s0 (should be ok)
- cb read x from s1 (should hang)
- cb write y to s1 (should be ok)
- cc write x to s1 (should be ok)
- cb read x from s1 (should be ok)
- cb write y to s0 (should be ok)
- cc read x from s0 (should hang)
- Partition Heals
- ca read x from s0 (should be ok)
- cc read x from x0 (should be ok)
- cc reads y from x0 (should be ok)

#### Explicit violations
A violation can be observed if any of the operations that are supposed to hang terminates, or any of the operations that is supposed to be ok hangs. Additionally, the correct and most updated value should be retrieved.

#### Lamport Diagram
![Lamport Diagram - Bob doesn't smell example](./lamport_causal_bob_doesnt_smell_old.jpg)

### Causal Tiebreak Basic
Tests tiebreaking and consistency in breaking ties between concurrent writes of the same key to two partitioned servers.
Based on [Evan's pseudocode](https://discord.com/channels/1326277272566501486/1340435708254617600/1346238343330926693).
```
partition(a)
partition(b)
c1.put(a, "x", 1)
c2.put(b, "x", 2)
if c1.get(b, "x") times out:
    c1_hang = True
if c2.get(a, "x") times out:
    c2_hang = True

partition(a, b)
xa = c3.get(a, "x")
xb = c4.get(b, "x")
assert xa == xb

if xa == 1:
    assert c1_hang and not c2_hang
if xa == 2:
    assert c2_hang and not c1_hang
```

#### Explicit Violations
If both or neither request hangs when c1 and c2 make a GET during the partition. If two other clients do not agree when reading the same value, the tiebreak was not consistent.

#### Lamport Diagram
![Lamport diagram - causal tiebreak basic](./lamport_causal_basic_tiebreak.jpg)

### Causal advanced: Bob doesn't smell(new - updated 3/2/2025)

#### Test Description
This test involves 3 clients and 2 server nodes, and is a play on the traditional bob smells test. The test involves two stages, the first, where the two servers are partitioned, and the second, where the two servers can talk to each other. From now on, we will refer to the clients as ca, cb, cc, and the servers as s0, s1. The only possible keys are x and y

The procedure is as follows:

- ca put x in s0 (should be ok)
- ca read x from s1 (should hang)
- cb read x from s0 (should be ok)
- cb read x from s1 (should hang)
- cb write y to s1 (should be ok)
- cc reads y from s1(should be ok)
- cc reads x from s0(should be ok)
- cc write x to s1 (should be ok)
- cb read x from s1 (should not hang to preserve avaliability, assuming the previous execution is successful)
- cb write y to s0 (should be ok)
- cc read x from s0 (should hang)
- *Partition Heals*
- ca read x from s0 (should be ok)
- cc read x from s0 (should be ok)
- cc reads y from s0 (should be ok)

#### Explicit violations
A violation can be observed if any of the operations that are supposed to hang terminates, or any of the operations that is supposed to be ok hangs. Additionally, the correct and most updated value should be retrieved.

#### Lamport Diagram
![Lamport Diagram - Bob doesn't smell example](./lamport_causal_bob_doesnt_smell.jpg)

### Causal basic: Get all keys (Updated 3/3/2025)
This is a test that involves 3 clients and 2 server nodes and ensures causal consistency between two servers even though the servers are not able to communicate with each other. There are three components to this test. First, client 1 sends x to server 1, and y to server 2. Both of these operations are expected to succeed. Second, client 1 requests all from both server 1 and server 2, which are both supposed to hang, client 2 requests all from server 1, which is expected to succeed, and requests all from server 2, which is expected to hang. Finally, client 3 requests all from server 2, which is expected to pass. The validity of this test is discussed [here](https://discord.com/channels/1326277272566501486/1340435708254617600/1345999275322118228), *and updated [here](https://discord.com/channels/1326277272566501486/1340435708254617600/1346201166283018333)*. 

This test is updated on 3/3/2025 after reflecting on correctness if client 3 should hang or not. As the test contributers currently understand it, the client should not hang, as it is equivalent to executing parallel gets, with no causal dependencies on each other. 

#### Explicit violations
A violation can be observed if any of the operations that are supposed to hang terminates, or any of the operations that is supposed to be ok hangs. Additionally, the correct value should be retrieved in the case that is successful.

#### Lamport Diagram
![Lamport Diagram - Get all keys 1 example](./lamport_causal_get_all_1.jpg)

### Causal put no update 1-2
We note that in causal relationship, *reads* establish the existence of the ordering(a user can always write, but some reads may hang). This test is a twist on the traditional bob smells fua test, but instead of establishing a hang after readying y, then reading x, we note that reading x as a new client allows a 404.

The only difference between 1, 2 is the existence of data in server 2 before the initial bob smells execution

Updated 3/3/2025: The test is updated to accurately reflect the lamport diagram, the only difference is the order of write x in server 1 and server 2.


#### Explicit Violations
Hanging on places the server is not supposed to, or termination when the server is supposed to hang.

#### Lamport Diagram
![Lamport Diagram - Bob smells read](./bob_smells_read.jpg)
![Lamport Diagram - Bob smells read 2](./bob_smells_read_2.jpg)

### Causal no overwrite future

This is based on a bug I had in an earlier version of my kvs. On read, the server would completely overwrite the metadata without properly merging it.
This causes the client to be able to read versions of a key from the past.

Client writes y to A, and then writes x to A, which means that x depends on the first y write. Then A and B are partitioned. Client writes y again,
which happens after the first write to y. Client reads x from B, which should immediately answer with the value of x. Client then tries to read
y from B, which should hang, because the client knows about a future version of y from after the partition.

#### Explicit Violations
The server returns the version of x from the past (client knows about future version).

#### Lamport Diagram
![Lamport Diagram - no overwrite future](./no_overwrite_future.jpg)
