# Eventual Consistency

## Eventual Consistency Basic

### Convergence Concurrent Basic 1

#### Test Description
This is a basic test that has 4 clients and 3 servers, with no partitions. 3 clients perform exactly one write to a unique server, with the 4th client reading the value from all 3 servers 10 seconds (which can be modified) after the final write. The requests are concurrent and the time limit for eventual consistency has elapsed, so any value written to a server is valid, as long as all three servers agree.

#### Explicit Violation
A violation of eventual consistency, in which at least two servers disagree about the value of x.

#### Lamport Diagram
![Lamport Diagram - Convergence Concurrent Basic 1](./lamport-convergence-concurrent-basic-1.png)

### Convergence Happens Before Basic 1

#### Test Description
A single client performs 3 sequential writes to 3 unique servers. After the writes complete, the same client reads from each server, and should get the same result from each server.

#### Explicit Violation
The client recieves different values of x from any 2 servers.

#### Lamport Diagram
![Lamport Diagram - Convergence Happens Before Basic 1](./lamport-convergence-happens-before-basic-1.png)

### Convergence Get All 1 (basic)

#### Test Description
A basic test for convergence using the `GET /data/` endpoint. We have 2 clients and 2 servers. The order of opertions is:

1. Partition({A})
2. Partition({B})
3. c1.PUT(A, x=alice)
4. c1.PUT(B, z=testing123)
5. c2.PUT(B, x=bob)
6. c2.PUT(B, y=builder)
7. c2.GETALL(A), expect 408/timeout 
8. Partition({A, B}), Sleep(10)
9. c2.GETALL(A)
10. c2.GETALL(B)
11. c1.GETALL(A)
12. c1.GETALL(B)

The test will check if the states of the final 4 GETALL calls are equivalent to check for properly converging values for both servers.

#### Explicit violation
If the clients timeout during reads, failed PUTs, or inconsistent final KV store states.

#### Lamport Diagram
![Lamport Diagram - Convergence Get All 1](./lamport-convergence-get-all1.jpg)

### Convergence Get All 2 (advanced)

#### Test Description
The test starts with three servers aware of each other yet unable to communicate. There are three writer clients and three reader clients. The writer clients write to the three servers differently, with the same keys and different values. The reader clients are then reads the data from the three servers separately. Finally, the partition heals and the servers can communicate with each other again. The reader clients then read the data from the servers again and ensures that the data is consistent.

#### Explicit Violation
The reader clients hang in any reads, or the data is inconsistent in the end.

#### Lamport Diagram
![Lamport Diagram - Convergence Get All2 ](./lamport-convergence-get-all.jpg)
